"""Effective Dimensionality Reduction"""

import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.utils import check_array
from sklearn.base import clone
from sklearn.utils.validation import check_is_fitted
from .base import BaseEDR


class EffectiveDimensionalityReduction(BaseEDR):
    """Effective dimensionality reduction class

    Parameters
    ----------
    estimator : object
        A supervised learning estimator with a ``predict_gradient``
        and ``fit`` methods.
    dr_transformer : object
        A linear dimensionnality reduction method that provides
        information about new axes through ``components_`` attribute.
    n_components : int (default=None)
        Number of components to left after fitting. If None then 
        n_components = n_features
    step : int, float (default=None)
        Number of components to drop at each iteration. If step is float
        then number of components to drop at each iteration defines as 
        number of worst components with sum of subspace variance lower then 
        1 - step. If step is None only one iteration is applied.
    normalize : bool, optional (default=True)
        If True input data will be normalized by ``StandardScaler``
        before fitting.
    preprocessor : object, optional (default=None)
        Performs preprocessing of input data if it was normalized.
        If None preprocessing won't be done. Use PCA, SparsePCA etc.

    Attributes
    ----------
    components_ : array, shape (n_components, n_features)
        New axes in feature space, representing the directions of
        maximum variance of the target.
        n_components = ``dr_transformer.n_components``
    feature_importances_ : ndarray, shape (n_components, n_features)
        The contribution of each feature to new axes.
        n_components = ``dr_transformer.n_components``
    estimator_ : object
        Estimator fitted to preprocessed data.
    dr_transformer_ : object
        Dr_transformer fitted to gradients.
    scaler_ : object
        ``StandardScaler`` fitted to raw data.
    preprocessor_ : object
        Preprocessor fitted to normalized data.
    subspace_variance_: array, shape (n_components, )
        Subspace variance calculated as tr(X.T * X) - tr(Y_i.T * Y_i)
        where Y_i=XU_i, U_i - orthogonal complement for components_.T[:, :i],
        i =  1, ..., n_components
        n_components = ``dr_transformer.n_components``
    subspace_variance_ratio_: array, (n_components, )
        Subspace variance ratio calculated as subspace_var_/tr(X.T * X)
        n_components = ``dr_transformer.n_components``

    If `refit` method has been applied the following attributes
    are also presented:

    refit_transformer_ : object
        `refit_transformer` fitted on gradients estimated during `fit`.
        Attribute is present only if the `refit` has been applied.
    refit_components_ : array, shape (n_refit_components, n_features)
        New axes in feature space, representing the directions of
        maximum variance of the target.
        n_components = ``refit_transformer.n_components``
    refit_subspace_variance_: array, shape (n_components, )
        Subspace variance calculated as tr(X.T * X) - tr(Y_i.T * Y_i)
        where Y_i=XU_i, U_i - orthogonal complement for components_.T[:, :i],
        i =  1, ..., n_components
        n_components = ``refit_transformer.n_components``
    refit_subspace_variance_ratio_: array, (n_components, )
        Subspace variance ratio calculated as subspace_var_/tr(X.T * X)
        n_components = ``refit_transformer.n_components``
    """

    def __init__(self, estimator, dr_transformer, n_components=None, step=None,
                 normalize=True, preprocessor=None):
        self.normalize = normalize
        self.preprocessor = preprocessor
        super(EffectiveDimensionalityReduction, self).__init__(
            estimator, dr_transformer, n_components, step)

    def fit(self, X, y=None, **opt_kws):
        """Fit the model with X, y

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples in the number of samples
            and n_features is the number of features.
        y : array-like, shape (n_samples,)
            Target values.

        Returns
        -------
        self : object
            Returns self.
        """
        X = self._preprocessing_fit(X)
        super(EffectiveDimensionalityReduction,
              self).fit(X, y, **opt_kws)
        if self.normalize is True:
            self.components_ = np.dot(self.components_, self._reverse_scaling_)
            if hasattr(self, 'refit_components_'):
                self.refit_components_ = np.dot(self.refit_components_,
                                                self._reverse_scaling_)
        return self

    def _preprocessing_fit(self, X, transform=True):
        """Preprocess X with ``StandardScaler`` and `preprocessor`

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples in the number of samples
            and n_features is the number of features.
        transform : bool, optional (default=True)

        Returns
        -------
        X_new : ndarray, shape (n_samples, n_componets)
        """
        if not self.normalize:
            if self.preprocessor is not None:
                mes = 'To apply prerpocessing, normalize should be True'
                raise ValueError(mes)
            return X
        self.scaler_ = StandardScaler()
        X_preprocessed = self.scaler_.fit_transform(X)
        self._scaling_ = np.diag(self.scaler_.scale_)
        self._reverse_scaling_ = np.diag(1 / self.scaler_.scale_)
        # note that X will be centered during training to improve
        # robustness of GP models.
        # the transform step will be a pure linear map without a translation

        if self.preprocessor is not None:
            self.preprocessor_ = clone(self.preprocessor)
            X_preprocessed = self.preprocessor_.fit_transform(X_preprocessed)
            self._check_transformer(self.preprocessor_)
            # save preprocesing map
            self._preprocessing_ = self.preprocessor_.components_

        return X_preprocessed if transform else None

    def _preprocessing_transform(self, X):
        """Preprocess new data

        Applies preprocessing steps as to data in ``fit``

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)

        Returns
        -------
        X_new : ndarray, shape (n_samples, n_componets)
        """
        X = check_array(X)
        if self.normalize is True:
            check_is_fitted(self, 'scaler_')
            X = self.scaler_.transform(X)
        # if self.preprocessor is not None:
        #     check_is_fitted(self, 'preprocessor_')
        #     X = self.preprocessor_.transform(X)
            X = np.dot(X, self._scaling_)
        X = np.dot(X, self.components_.T)
        return X

    def get_estimator_gradients(self, X):
        """Returns gradients of the sample

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data for which gradients will be calculated.

        Returns
        -------
        grad : ndarray, shape (n_samples, n_features)
            Calculated gradients.
        """
        X = check_array(X)
        # X_transform = self.transform(X)
        # subsp_grads = self.estimator_.predict_gradient(X_transform)
        return self._get_estimator_gradients(X, True)

    def _get_estimator_gradients(self, X, preprocessing_transform=False):
        """Returns gradients of the sample

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data for which gradients will be calculated.
        preprocess : bool, optional (default=False)
                Whether to preprocess data or not.

        Returns
        -------
        grad : ndarray, shape (n_samples, n_features)
            Calculated gradients in the non-preprocessed space.
        """
        if preprocessing_transform:
            X = self._preprocessing_transform(X)
        check_is_fitted(self, 'estimator_')
        grad = self.estimator_.predict_gradient(X)
        if (self.preprocessor is not None and
            self.num_iter == 0 and
                not preprocessing_transform):
            check_is_fitted(self, 'preprocessor_')
            grad = np.dot(grad, self._preprocessing_)
        if preprocessing_transform:
            grad = np.dot(grad, self.components_)
        return grad

    @property
    def feature_importances_(self):
        """Return the feature importances of each feature in new axes.

        Feature importances are calculated with respect to scaling.

        Returns
        -------
        feature_importances_ : ndarray, shape (n_components, n_features)
            The contribution of each feature to new axes.
            n_components = ``dr_transformer.n_components``
        """
        check_is_fitted(self, 'components_')
        importances_ = self.components_
        if self.normalize is True:
            importances_ = np.dot(importances_, self._scaling_)
        return importances_

    def transform(self, X, refitted=False):
        check_is_fitted(self, 'components_')
        X = check_array(X)
        if refitted:
            check_is_fitted(self, ['refit_transformer_', 'refit_components_'])
            return np.dot(X, self.refit_components_.T)
        if hasattr(self, '_gradients_') and self._gradients_ is not None:
            components = self.components_
        else:
            components = (self.components_ if self.preprocessor is None else
                          np.dot(self.components_, self._preprocessing_.T))
        return np.dot(X, components.T)



import numpy as np
from scipy.stats import special_ortho_group
from sklearn.utils import check_array
from math import pi


def get_gaussian_inputs(sample_size, eig_values, eig_vectors=None, mean=None):
    # generate covariance mat
    dim = len(eig_values)
    eig_values = np.diag(eig_values)
    if eig_vectors is None:
        eig_vectors = special_ortho_group.rvs(dim)
    else:
        eig_vectors = check_array(eig_vectors)
        if eig_vectors.shape != (dim, dim):
            raise ValueError('eig_vectors shape must be ({0},{0})'.format(dim))
    cov = np.dot(np.dot(eig_vectors, eig_values), eig_vectors.T)
    # generate centered inputs
    if mean is None:
        mean = np.zeros(dim)
    X = np.random.multivariate_normal(mean, cov, sample_size)
    return X


def get_tanh_targets(X, coefs, bias=0, noise_std=0.05):
    if X.shape[1] != len(coefs):
        raise ValueError(
            'Dimensionality of input ({}) and coefs ({}) are mismatched')
    y = np.tanh(np.dot(X, coefs) + bias)
    y += noise_std * np.random.randn(X.shape[0])
    return y


def get_beta_inputs(sample_size, ndim, tau=1):
    "Generates data from 2*Beta(1, tau)-1"
    return 2*np.random.beta(1, tau, size=(sample_size, ndim))-1


def get_edr_target(X, sigma=None):
    """"Generates target for different dimensions

    n=1 => g(u) = u sin(sqrt(5) u)
    n=2 => g(u1, u2) = (u1^3+u2)(u1-u2^3)
    n=3 => g(u1, u2, u3) = (u1^3+u2)(u1-u2^3)+u3*u3_coef
    """
    effective_ndim = X.shape[1]

    if effective_ndim == 1:
        result = np.array([u*np.sin(np.sqrt(5)*u) for u in X])
    if effective_ndim == 2:
        result = np.array([(u1 ** 3+u2)*(u1-u2 ** 3) for u1, u2 in X])
    if effective_ndim == 3:
        result = np.array([(u1 ** 3+u2)*(u1-u2 ** 3)+u3 for u1, u2, u3 in X])
    result = result.ravel()
    if sigma is not None:
        result += sigma * np.random.randn(result.size)
    return result


def get_branin_targets(X, noise_std=None):
    '''Branin function generator
    It's a smooth 2D function which is frequently used as a target for
    testing of optimization algorithms.
    For details see https://www.sfu.ca/~ssurjano/branin.html

    Parameters
    ----------
    X : np.array
        The input data. There are two requirements:
         - must be a 2D array with a shape [n_samples, 2]
         - both columns should be in range [0, 1]
    noise_std : float or None
        Standard deviation of iid Gaussian noise which will be added to outputs

    Returns
    -------
    y : np.array
        Vector of target, [n_samples, ]
    '''
    # ToDo: check X

    # params
    a, b, c = 1, 5.1 / (4 * pi**2), 5 / pi
    r, s, t = 6, 10, 1 / (8 * pi)
    # x0 ∈ [-5, 10], x1 ∈ [0, 15].
    x0 = 15 * X[:, 0] - 5
    x1 = 15 * X[:, 1]
    y = a * (x1 - b * x0**2 + c * x0 - r)**2 + s * (1 - t) * np.cos(x0) + s
    if noise_std is not None:
        y += noise_std * np.random.randn(X.shape[0])
    return y



"""Base class for Effective Dimensionality Reduction"""

import numpy as np
from copy import deepcopy
from sklearn.base import TransformerMixin, clone
from sklearn.utils import check_array, check_X_y
from sklearn.utils.validation import check_is_fitted
from .utils import subspace_variance_ratio


class BaseEDR(TransformerMixin):
    """Base class for effective dimensionality reduction

    Parameters
    ----------
    estimator : object
        A supervised learning estimator with a ``predict_gradient``
        and ``fit`` methods.
    dr_transformer : objec
        A linear dimensionnality reduction method that provides
        information about new axes through ``components_`` attribute.
    n_components : int (default=None)
        Number of components to left after fitting. If None then 
        n_components = n_features
    step : int, float (default=None)
        Number of components to drop at each iteration. If step is float
        then number of components to drop at each iteration defines as 
        number of worst components with sum of subspace variance lower then 
        1 - step. If step is None only one iteration is applied.

    Attributes
    ----------
    components_ : array, shape (n_components, n_features)
        New axes in feature space, representing the directions of
        maximum variance of the target.
        n_components = ``dr_transformer.n_components``
    feature_importances_ : ndarray, shape (n_components, n_features)
        The contribution of each feature to new axes.
        n_components = ``dr_transformer.n_components``
    estimator_ : object
        Estimator fitted to preprocessed data.
    dr_transformer_ : object
        Dr_transformer fitted to gradients.
    subspace_variance_: array, shape (n_components, )
        Subspace variance calculated as tr(X.T * X) - tr(Y_i.T * Y_i)
        where Y_i=XU_i, U_i - orthogonal complement for components_.T[:, :i],
        i =  1, ..., n_components
        n_components = ``dr_transformer.n_components``
    subspace_variance_ratio_: array, (n_components, )
        Subspace variance ratio calculated as subspace_var_/tr(X.T * X)
        n_components = ``dr_transformer.n_components``

    If `refit` method has been applied the following attributes
    are also presented:

    refit_transformer_ : object
        `refit_transformer` fitted on gradients estimated during `fit`.
        Attribute is present only if the `refit` has been applied.
    refit_components_ : array, shape (n_refit_components, n_features)
        New axes in feature space, representing the directions of
        maximum variance of the target.
        n_components = ``refit_transformer.n_components``
    refit_subspace_variance_: array, shape (n_components, )
        Subspace variance calculated as tr(X.T * X) - tr(Y_i.T * Y_i)
        where Y_i=XU_i, U_i - orthogonal complement for components_.T[:, :i],
        i =  1, ..., n_components
        n_components = ``refit_transformer.n_components``
    refit_subspace_variance_ratio_: array, (n_components, )
        Subspace variance ratio calculated as subspace_var_/tr(X.T * X)
        n_components = ``refit_transformer.n_components``

    """

    def __init__(self, estimator, dr_transformer, n_components=None,
                 step=None):
        self.estimator = estimator
        self.dr_transformer = dr_transformer
        self.n_components = n_components
        self.step = step

    def fit(self, X, y=None, **opt_kws):
        """Fit the model with X, y

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples in the number of samples
            and n_features is the number of features.
        y : array-like, shape (n_samples,)
            Target values.

        Returns
        -------
        self : object
            Returns self.
        """

        if self.n_components is None:
            self.n_components_ = X.shape[1]
            if isinstance(self.step, int):
                mes = "If step is int n_components must be specified"
                raise ValueError(mes)
        else:
            self.n_components_ = self.n_components

        self.adaptive_step = False
        if self.step is None:
            self.step_ = self.n_components_
        elif isinstance(self.step, int) and self.step > 0:
            self.step_ = self.step
        elif isinstance(self.step, float) and 0 < self.step < 1:
            if self.n_components is not None:
                mes = "If step is float n_components should be None"
                raise ValueError(mes)
            self.adaptive_step = True
            self.step_ = self.step
        else:
            mes = "Step should be None or int > 0 or float from 0 to 1"
            raise ValueError(mes)

        self._gradients_ = None
        self.components_ = None
        self.continue_iteration = True
        self.num_iter = 0
        X_proj = X.copy()
        while self.continue_iteration:
            self._fit_estimator(X_proj, y, **opt_kws)
            self._fit_dr_transformer(X_proj)
            X_proj = self.transform(X)
            self.num_iter += 1

        self._last_fit(X_proj, y, **opt_kws)
        return self

    def refit(self, refit_transformer):
        """Compute new components using gradients estimated during fit.

        It uses gradients estimated during fit and finds the right subspace 
        for them using `refit_transformer`. To transform data with 
        found components use `transform` with `refiitted=True`

        Parameters
        ----------
        refit_transformer : object
            Transformer to fit with gradients estimated on fit.
            It should have attribute `components_` after fit.

        Returns
        -------
        self : object
            Returns self.
        """
        check_is_fitted(self, 'components_')
        self.refit_transformer_ = clone(refit_transformer)
        self.refit_transformer_.fit(self._gradients_)
        self._check_transformer(self.refit_transformer_)
        self.refit_components_ = deepcopy(self.refit_transformer_.components_)
        self.refit_components_ = (
            self.refit_components_/np.linalg.norm(self.refit_components_,
                                                  axis=1).reshape(-1, 1))
        (self.refit_subspace_variance_,
         self.refit_subspace_variance_ratio_) = subspace_variance_ratio(
            self._gradients_,
            self.refit_components_.T)
        return self

    def _last_fit(self, X, y, **opt_kws):
        """Compute gradients for original and effective subspaces.

        Also computes subspace variance ratio for gradients in effective 
        subspace with respect to original gradients

        Parameters
        ----------
        X : array-like, shape (n_samples, n_componets)
            Training data, projected to effective subspace.
        y : array-like, shape (n_samples,)
            Target values.

        Returns
        -------
        self : object
                Returns self.

        """
        self._fit_estimator(X, y, **opt_kws)
        check_is_fitted(self, 'estimator_')
        grad = self._get_estimator_gradients(X)
        self.subspace_gradients_ = grad
        self._gradients_ = np.dot(grad, self.components_)
        (self.subspace_variance_,
         self.subspace_variance_ratio_) = subspace_variance_ratio(
            self._gradients_,
            self.components_.T)
        return self

    def _fit_estimator(self, X, y, **opt_kws):
        """Fit the estimator with X, y.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples in the number of samples
            and n_features is the number of features.
        y : array-like, shape (n_samples,)
            Target values.

        Returns
        -------
        self : object
            Returns self.
        """
        X, y = check_X_y(X, y, accept_sparse=False)
        if y is not None:
            self.estimator_ = clone(self.estimator)
            self.estimator_.fit(X, y, **opt_kws)
        elif not hasattr(self, 'estimator_'):
            self.estimator_ = clone(self.estimator)
            # we will check later that the estimator is properly fitted
        return self

    def _fit_dr_transformer(self, X):
        """Fit the transformer with X and calculate subspace_var_ and
        subspace_var_ratio_

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples in the number of samples
            and n_features is the number of features.

        Returns
        -------
        self : object
            Returns self.
        """
        check_is_fitted(self, 'estimator_')

        grad = self._get_estimator_gradients(X)

        self.dr_transformer_ = clone(self.dr_transformer)
        self.dr_transformer_.fit(grad)
        self._check_transformer(self.dr_transformer_)
        components = deepcopy(self.dr_transformer_.components_)

        if self.adaptive_step:
            _, var_ratio_ = subspace_variance_ratio(grad, components.T)
            n_components = np.sum(np.cumsum(var_ratio_) < self.step_,
                                  dtype=int) + 1
            if n_components == grad.shape[1]:
                self.continue_iteration = False
        else:
            n_components = max(self.n_components_, X.shape[1] - self.step_)
            if n_components == self.n_components_:
                self.continue_iteration = False

        components = components[:n_components, :]
        self.components_ = (components if self.components_ is None else
                            np.dot(components, self.components_))
        return self

    def get_estimator_gradients(self, X):
        """Returns gradients of the sample

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data for which gradients will be calculated.

        Returns
        -------
        grad : ndarray, shape (n_samples, n_features)
            Calculated gradients.
        """
        X = check_array(X)
        return self._get_estimator_gradients(X)

    def _get_estimator_gradients(self, X):
        """Returns gradients of the sample and check model is fitted

        See ``get_estimator_gradients``
        """
        check_is_fitted(self, 'estimator_')
        grad = self.estimator_.predict_gradient(X)
        return grad

    def transform(self, X, refitted=False):
        """Apply dimensionality reduction on X.

        X is projected on the components previous extracted
        from a training set.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            New data, where n_samples in the number of samples
            and n_features is the number of features.
        refitted : bool (False)
            Whether to transform using refit_transformer components.
            May be set to `True` only if `refit` was applied before
        Returns
        -------
        X_new : ndarray, shape (n_samples, n_components)
        """
        check_is_fitted(self, 'components_')
        X = check_array(X)
        if refitted:
            check_is_fitted(self, ['refit_transformer_', 'refit_components_'])
            return np.dot(X, self.refit_components_.T)
        return np.dot(X, self.components_.T)

    def inverse_transform(self, X):
        """Transform data back to its original space.

        Returns an array X_original whose transform would be X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_components)
            New data, where n_samples in the number of samples
            and n_components is the number of components.

        Returns
        -------
        X_original : ndarray, shape (n_samples, n_features)
        """
        check_is_fitted(self, 'components_')
        X = check_array(X)
        return np.dot(X, np.linalg.pinv(self.components_).T)

    @property
    def feature_importances_(self):
        """Return the feature importances of each feature in new axes.

        Returns
        -------
        feature_importances_ : ndarray, shape (n_components, n_features)
            The contribution of each feature to new axes.
            n_components = ``dr_transformer.n_components``
        """
        check_is_fitted(self, 'components_')
        return self.components_

    def _check_transformer(self, transformer):
        """Check that transformer has attribute ``components_``.

        Parameters
        ----------
        transformer : object
        """
        if not hasattr(transformer, 'components_'):
            raise AttributeError('The transformer does not expose '
                                 '"components_" attribute')


class ExtendedEDR(BaseEDR):

    def __init__(self, estimator, dr_transformer, refit_transformer,
                 n_components=None, step=None):
        super(ExtendedEDR, self).__init__(estimator, dr_transformer,
                                          n_components, step)
        self.refit_transformer = refit_transformer

    def fit(self, X, y=None, **opt_kws):
        super(ExtendedEDR, self).fit(X, y, **opt_kws)
        self.refit(self.refit_transformer)

    def transform(self, X):
        super(ExtendedEDR, self).transform(X, refitted=True)



# Always prefer setuptools over distutils
from setuptools import setup
# To use a consistent encoding
from codecs import open
from os import path
from pip.req import parse_requirements

# read dependencies from requirements.txt
install_reqs = parse_requirements('requirements.txt', session=False)
reqs = [str(ir.req) for ir in install_reqs]


here = path.abspath(path.dirname(__file__))
# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()
description = 'Effective Dimensionality Reduction based on Gaussian Processes'

setup(
    name='edr-gp',
    version='0.2.1',
    description=description,
    long_description=long_description,
    url='https://github.com/neuro-ml/edr-gp',
    author='Neuro.ml group',
    author_email='belyaevmichel+edr-gp@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.5',
    ],
    packages=['edrgp'],
    install_requires=reqs,
)


import scipy
import numpy as np
from copy import deepcopy
from sklearn.utils import check_array
from sklearn.base import TransformerMixin, BaseEstimator


def ort_space(A):
    """Retutn orthogonal space for matrix a

    Parametrs
    ---------
    A: np.ndarray
        The matrix [n_features, n_components] for which we want to find
        orthogonal space

    Returns
    -------
    U: np.ndarray
        The matrix [n_features, n_features - n_components] which columns are
        basis in orthogonal space
    """
    U, s, _ = scipy.linalg.svd(a=A, full_matrices=True)
    return U[:, sum(abs(s) > 1e-10):]


def subspace_variance_ratio(X, V):
    """Compute subspace variance for V in X

    If V is not orthonormalized then it will return only one 
    value: subspace variance for the whole matrix V.

    Parameters
    ----------
    X : array, shape (n_samples, n_features)

    V : array, shape (n_feaures, n_components)
        Projector components

    Returns
    -------
    subspace_variance_ : array, shape (n_components, )
        Subspace variance for each column of V (or for the whole V 
        if it is not orthonormalized)
    subspace_variance_ratio_ : array, shape (n_components, )
        Subspace variance ratio for each column of V
    """
    if np.allclose(np.dot(V.T, V), np.eye(V.shape[1])):
        subspace_variance_ = np.linalg.norm(X.dot(V), axis=0)
    else:
        V_orthonormalized = np.linalg.qr(V)[0]
        subspace_variance_ = np.linalg.norm(X.dot(V_orthonormalized))

    subspace_variance_ratio_ = (subspace_variance_/np.linalg.norm(X)) ** 2
    return subspace_variance_, subspace_variance_ratio_


def discrepancy(B, V):
    """Measure of difference between two subspaces

    The discrepancy is measuared by the following formula:
    ||BB^T(I - VV^T)||_F/d, where d denotes number of true components and
    I is an identity matrix with shape n_features.

    Parameters
    ----------
    B : array, shape (n_features, n_components_true)
        The true projector on subspace

    V : array, shape (n_features, n_components)
        The estimated projector on subspace

    Returns
    -------
    """
    return np.linalg.norm(np.dot(B.dot(B.T),
                                 (np.eye(B.shape[0]) - 
                                                    V.dot(V.T)))) / B.shape[1]


class CustomPCA(BaseEstimator, TransformerMixin):
    """PCA without centering and scaling

    Linear dimensionality reduction using Singular Value 
    Decomposition of the data to project it to a lower dimensional 
    space.

    Parameters
    ----------
    n_components : int, float, None or string
        Number of components to keep.
        if n_components is not set all components are kept::

            n_components == min(n_samples, n_features)

        if ``0 < n_components < 1``, select the number
        of components such that the amount of variance that needs to be
        explained is greater than the percentage specified by n_components.

    Attributes
    ----------
    components_ : array, shape (n_components, n_features)
        Principal axes in feature space, representing the directions of
        maximum variance in the data. The components are sorted by
        ``explained_variance_``.

    subspace_variance_ : array, shape (n_components,)
        The amount of variance contained in each of the selected components.

        Equal to n_components largest eigenvalues
        of the covariance matrix of X.

    subspace_variance_ratio_ : array, shape (n_components,)
        Percentage of variance contained in each of the selected components.

        If ``n_components`` is not set then all components are stored and the
        sum of subspace variances is equal to 1.0.
    """

    def __init__(self, n_components=None):
        self.n_components = n_components

    def fit(self, X, y=None):
        """Fit the model with X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples in the number of samples
            and n_features is the number of features.

        y : Ignored.

        Returns
        -------
        self : object
        """
        X_copy = check_array(X, copy=True)

        U, S, Vh = np.linalg.svd(X_copy)
        subspace_var_ratio = S**2 / np.sum(S**2)

        if self.n_components is None:
            n_components = X_copy.shape[1]
        if (isinstance(self.n_components, int)
                and 0 < self.n_components <= X_copy.shape[1]):
            n_components = self.n_components
        if isinstance(self.n_components, float) and 0 < self.n_components < 1:
            n_components = np.sum(np.cumsum(expl_var_ratio) <
                                  self.n_components, dtype=int) + 1
        n_components = min(X_copy.shape[0], n_components)

        self.components_ = Vh[:n_components, :]
        self.subspace_variance_ = (S**2)[:n_components]
        self.subspace_variance_ratio_ = subspace_var_ratio[:n_components]

        return self

    def transform(self, X):
        """Apply dimensionality reduction to X.

        X is projected on the first principal components previously 
        extracted from a training set.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            New data, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)
        """
        return X.dot(self.components_.T)
